#------------------------------------------------------------------------------------------------
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#------------------------------------------------------------------------------------------------
#    Name:	VolumeBySlice
#      By:	sdmitch
#   Usage:	Compute the volume by slices of a given selected "solid" 3D shape.
#	 	 Note:	The front and back faces are assummed to be vertical.  If there are other vertical faces, 
#						deselect them before running the plugin.  Only edges that connect to the "front" face 
#						vertices are used.  Hidden and soft edges are ignored.  The area and volume data are 
#						written to a text file.
#    Date:	Jun 2011
#------------------------------------------------------------------------------------------------
#
unless file_loaded?(__FILE__)
    UI.menu('Plugins').add_item('Volume By Slice') { VBS.calc_vol }
	file_loaded(__FILE__)
end
#
module VBS

	def VBS.calc_vol
		mod=Sketchup.active_model
		ent=mod.entities
		sel=mod.selection
		edges=[]; faces=[]; face=nil
		# find the front and back faces
		for e in sel
			if e.is_a? Sketchup::Face
				norm = e.normal; puts "norm = #{norm}"
				faces.push(e) if norm.z.abs < 0.000001
			end
		end
		if (faces.length != 2)
			puts "faces.length = #{faces.length}"
			mod.start_operation "paint faces"
			for f in faces
				f.material = [255,0,0]
				f.back_material = [255,0,0]
			end
			UI.messagebox "deselect the non front and back faces"
			mod.commit_operation
			return nil
		end
		fpts=[]; lines=[];
		fpts=faces[0].outer_loop.vertices.collect {|v| v.position}
		# find the connecting edges
		for e in sel
			if e.is_a? Sketchup::Edge
				if (!e.hidden? && !e.soft?)
					fa=e.faces
					if (fa)
						tryme=true
						for i in 0...fa.length
							tryme=false if faces.include? fa[i]; # ignore edges of front or back face
						end
					end
					if tryme; # match edge to face vertex
						for p in fpts
							pt,vec=e.line; # try e.start.position
							if fpts.include? pt
								edges.push e; # save edge and line
								i = fpts.index(pt); lines[i] = [pt,vec]
								break; # exit for
							else; # try e.end
								pt = e.end.position; vec.reverse!
								if fpts.include? pt
									edges.push e; # save edge and line
									i = fpts.index(pt); lines[i] = [pt,vec]
									break; # exit for
								end
							end
						end
					end
				end
			end
		end
		puts "fpts.length=#{fpts.length}, edges.length=#{edges.length}"
		if edges.length != fpts.length
			UI.messagebox "Face vertices and edges counts don't match"
			return nil
		end
		pt=faces[0].bounds.center; vec=faces[0].normal; pln=faces[1].plane
		poi=Geom.intersect_line_plane([pt,vec],pln); norm=pt.vector_to(poi)
		@vol_max=pt.distance(poi)
		#	0=inches,1=feet,2=millimeters,3=centimeters,4=meters
		unit=mod.options["UnitsOptions"]["LengthUnit"]
		case unit # inches per unit conversion factors
			when 0;ipu = 1.0;units="inches";@vol_max = 100.0.inch if !@vol_max; @vol_spc=1.0.inch if !@vol_spc
			when 1;ipu = 12.0;units="feet";@vol_max = 100.0.feet if !@vol_max; @vol_spc=1.0.feet if !@vol_spc
			when 2;ipu = 0.03937007874;units="mm";@vol_max = 100.0.mm if !@vol_max; @vol_spc=1.0.mm if !@vol_spc
			when 3;ipu = 0.3937007874;units="cm";@vol_max = 100.0.cm if !@vol_max; @vol_spc=1.0.cm if !@vol_spc
			when 4;ipu = 39.37007874;units="m";@vol_max = 100.0.m if !@vol_max; @vol_spc=1.0.m if !@vol_spc
		end
		@rptdir = "c:/users/public/temp/"; @rptfile=faces[0].layer.name
		ans=UI.inputbox(["Report File Name:","Maximum Distance:","Slice Spacing:"],[@rptfile,@vol_max,@vol_spc],"Volume Calculation( "+units+"³ )")
		if !ans
			Sketchup.send_action "selectSelectionTool:"
			return nil
		end
		@rptfile=ans[0].to_s;@vol_max = ans[1].to_l; @vol_spc = ans[2].to_l; tot_vol=0.0
		vol_rpt = UI.savepanel("Volume Report",@rptdir,@rptfile+".txt")
		return nil if !vol_rpt
		fo=File.open(vol_rpt,"w")
		fo.printf("Volume Report for %s\n\n" , faces[0].layer.name)
		fo.printf("%5d%s Face Area: %.3f %s²\n",0,units,faces[0].area/ipu**2,units)
		mod.start_operation "volume"
		for i in 1..(@vol_max/@vol_spc).floor
			pts=[];pt=faces[i-1].bounds.center.offset(norm,@vol_spc); plane=[pt,norm]; 
			for j in 0...fpts.length
				pts.push Geom.intersect_line_plane(lines[j],plane)
			end
			faces[i] = ent.add_face(pts)
			vol = (((faces[i-1].area + faces[i].area)/2.0) * @vol_spc) / ipu**3
			fo.printf("        Volume: %.3f %s³\n" ,vol,units)
			fo.printf("%5d%s Face Area: %.3f %s²\n",i*@vol_spc/ipu,units,faces[i].area/ipu**2,units)
			tot_vol += vol
		end
		mod.commit_operation
		fo.printf("\nTotal Volume: %.3f %s³\n\n", tot_vol, units)
		fo.close
		Sketchup.undo
	end
end
